/* global -Promise */
/* jshint undef: true, unused: true */
/* global require, module, Packages, org, java */
var _ = require("lodash");
var Objects = require("convenience").Objects;
var request = require("request");
var Promise = require("rsvp").Promise;

var StringUtilities = Packages.cattura.utils.string.StringUtilities;
var URI = java.net.URI;

var Util = require("lib/util");

/*
 * Cranium plugin logger. This is provided to us by the Cranium Plugin.
 */
var logger;

/*
 * Internal API version number.
 *
 * Cranium 1.4.x corresponds to API version 2, but API version 1 is still supported
 * for backwards compatibility.
 */
var API_VERSION = 2;

/*
 * Flow:
 *  - 1: User POST to Cranium: /unit/add?redirect_uri=<redirect URI>
 *  - 2: Cranium POST to CaptureCast: /cranium/hello
 *       - Input: {"redirect_uri", "api_version", "cranium_url"}
 *       - New flow is created with: {"flow_id", "redirect_url", "user_cranium_url", "cranium_request_ip"}
 *       - Output: {"flow_id"} if unit is compatible with requested api version, OR false if not
 *  - 3: Cranium Redirects User to CaptureCast /cranium/authorize?flow_id=<flow ID>
 *  - 4: User POST to CaptureCast /cranium/authorize?flow_id=<flow ID>
 *  - 5: CaptureCast determines actual URL of Cranium using <cranium request IP>
 *  - 6: CaptureCast POST to Cranium /unit/<unit id>
 *       - Input: {"access_token"}
 *       - Cranium takes CaptureCast access token, creates unit from flow,
 *         generates own access token and sends it back
 *       - Output: {"access_token"}
 *  - 7: CaptureCast Redirects User to <redirect URI>
 */
function setup (options)
{
    var server = options.server;
    var authorizeTemplate = _(options.authorizeTemplate.read().getVal()).template();
    var addCranium = options.addCranium;
    var greet = options.greet;

    logger = options.logger;

    var flows = {}; // Requests to connect with Cranium that need to be confirmed.

    var HttpHandler = org.webbitserver.HttpHandler;

    /*
     * Endpoint for Cranium POST to CaptureCast.
     *
     * User starts the flow on the Cranium side, Cranium will POST to here to
     * generate a "flow" object which contains the context for the auth flow.
     * If CaptureCast accepts the request, it will pass the generated flow object
     * to Cranium, and then Cranium will redirect the user to the authorization
     * page for the user's authorization.
     *
     * A GET to this endpoint returns success and a version number. Cranium
     * can use this to determine if CaptureCast supports the v2 flow. If a
     * GET to this endpoint returns anything else (an error response or error code)
     * then Cranium can use the v1 flow or fail entirely.
     */
    server.add("/cranium/hello", new HttpHandler({
        handleHttpRequest: function (req, res) {
            if (req.method().toLowerCase() != "post")
            {
                sendSuccess(res, {
                    version: API_VERSION
                });
                return;
            }

            var remoteAddress = Util.getIPAddress(req);
            var data = JSON.parse(req.body());
            java.lang.System.err.println(JSON.stringify(data));
            if (!objectHas(data, [/*"api_version",*/ "redirectUrl", "craniumUrl", "tokenCallbackUrl"]))
            {
                logger.warning("Rejected hello from cranium at " + remoteAddress + " because of invalid post body");
                sendError(res, "Invalid post body (missing expected keys)");
                return;
            }

            /*if (data.api_version < API_VERSION || data.api_version >= API_VERSION + 1)
            {
                logger.warn("Rejected hello from cranium at " + remoteAddress + " because of API version mismatch");
                sendError(res, "API major version mismatch (expected " + API_VERSION + ", got " + data.api_version + ")");
                return;
            }*/

            var flowId = makeToken();
            var flow = {
                // Flow ID, used to uniquely identify a flow. Is seen by the user, so is not sensitive.
                id: flowId,

                // Shared secret between Cranium and CaptureCast. Not seen by the user.
                sharedSecret: makeToken(),

                // Cranium URL reported by Cranium's request. May not be the actual URL used by CaptureCast to connect to Cranium.
                userCraniumUrl: data.craniumUrl,

                // Redirect URL specified by Cranium. When the user is redirected here, the auth process is considered finished.
                redirectUrl: data.redirectUrl,

                // URL (relative to Cranium's base URL) to POST to exchange tokens.
                tokenCallbackUrl: data.tokenCallbackUrl.replace(":flowId", flowId),

                // IP address of Cranium's request. The actual URL used to transmit data from CaptureCast to Cranium may be derived from this.
                // This requires an up-to-date cattura-nginx configuration which passes over the client IP address to CaptureCast.
                craniumRequestIp: remoteAddress
            };

            // Preauthorized means that the flow does not require user input, i.e. the user just adds the unit in Cranium and then
            // the flow proceeds without further input. This can be because CaptureCast trusts the specified Cranium instance.
            flow.isPreauthorized = isPreauthorized(flow);

            // Endpoint that Cranium will redirect the user to.
            // TODO: if the flow is preauthorized, maybe we don't need to do any more
            // redirection and can just spin off a new thread to complete the flow from here.
            var authorizeEndpoint = "cranium/authorize?flow_id=" + flowId;

            flows[flowId] = flow;
            logger.info("Established new auth flow for cranium at " + remoteAddress);
            sendSuccess(res, {
                flow: flow,
                redirect_url: authorizeEndpoint
            });
        }
    }));

    /*
     * Authorization endpoint for CaptureCast's user to authorize adding the
     * unit to Cranium.
     *
     * This is the same endpoint as used in the v1 flow, but it expects the parameter
     * "flow_id" to differentiate it from a v1 flow. This "flow_id" is generated by
     * the "hello" endpoint (above) and passed into Cranium, which will redirect here.
     *
     * The v2 flow supports "preauthorization" which is a way to automatically
     * authorize Cranium to connect to a unit without passing through this
     * authorization prompt. Typically, "preauthorization" would be configured
     * in a settings UI and could consist of e.g. preauthorizing Cranium at
     * a certain hostname.
     */
    server.add("/cranium/authorize", new HttpHandler({
        handleHttpRequest: function (req, res, control) {
            var flowId = req.postParam("flow_id") || req.queryParam("flow_id");
            var nonce = req.postParam("nonce");
            if (!flowId)
            {
                control.nextHandler();
                return;
            }

            var flow = flows[flowId];
            if (!flow)
            {
                sendErrorPage(res, flow, "Invalid or expired auth flow");
                return;
            }

            /*
             * If the nonce is present, it can be assumed the user progressed
             * the auth flow. Check the nonce for validity.
             */
            if (nonce && flow.nonce && flow.nonce != nonce)
            {
                sendErrorPage(res, flow, "Invalid or expired nonce");
                return;
            }

            /*
             * Either proceed in auth flow, or prompt user.
             * If we are prompting the user, generate a nonce and pass that into
             * the client. The client will then repeat the nonce (along with flow id)
             * to confirm the authorization came from this client.
             *
             * If the flow is preauthorized, don't bother the user and just proceed
             * with the flow.
             */
            if (flow.isPreauthorized || flow.nonce)
            {
                /*
                 * Request access token from Cranium using our authorization code.
                 * This also verifies that we can access Cranium from the unit end.
                 * If Cranium gives us the access token, we redirect back to Cranium
                 * to finish the flow.
                 *
                 * Generate the inbound (Cranium -> CaptureCast) request token
                 * and pass it over to Cranium. Cranium will give us the outbound
                 * (CaptureCast -> Cranium) token, which we will use to talk to it.
                 *
                 * If this POST returns successfully, then the unit object is created
                 * on Cranium's side.
                 *
                 * Response will contain:
                 *  - access_token: outbound access token
                 *  - redirect_url: url to redirect user to
                 */
                var inboundAccessToken = makeToken();

                /*
                 * Discover the "actual" URL of Cranium here.
                 *
                 * This may be different than user_cranium_url, which is the URL
                 * provided by the user, if the user initiated the flow from
                 * a URL that we can't access. We can use the request IP to
                 * try to determine the URL to use.
                 *
                 * The discovered URL is only used for communication between
                 * CaptureCast and Cranium. For redirecting the end-user back
                 * to Cranium we use user_cranium_url.
                 */
                discoverCraniumUrl(flow).then(function (craniumUrl) {
                    flow.craniumUrl = craniumUrl;
                    logger.info("Discovered cranium url: " + flow.craniumUrl);

                    request.post(Util.buildUrl(flow.craniumUrl, flow.tokenCallbackUrl), {
                        body: {
                            shared_secret: flow.sharedSecret,
                            access_token: inboundAccessToken
                        },
                        json: true
                    }, function (err, response, body) {
                        /*
                         * If we get an error, display error message to user.
                         */
                        var errorMessage = checkForError(err, response, body);
                        if (errorMessage)
                        {
                            sendErrorPage(res, flow, errorMessage);
                            return;
                        }

                        logger.info("Successfully authorized cranium at " + flow.craniumUrl);
                        var cranium = addCranium({
                            url: flow.craniumUrl,
                            outboundAccessToken: body.access_token, // Token used by unit to talk to Cranium.
                            inboundAccessToken: inboundAccessToken // Token used by Cranium to talk to unit.
                        });

                        /*
                         * Push initial status updates to Cranium
                         */
                        try
                        {
                            cranium.sayHello();
                            cranium.pushInitialUpdate();
                            greet(cranium);
                            
                            /*
                             * Redirect back to Cranium to finish the flow.
                             */
                            res.status(307).header("Location", Util.buildUrl(flow.userCraniumUrl, flow.redirectUrl)).end();
                        }
                        catch (exception)
                        {
                            sendErrorPage(res, flow, exception.message);
                        }
                    });
                });
            }
            else
            {
                flow.nonce = makeToken();
                res.content(authorizeTemplate({
                    flow: flow,
                    error: null
                })).end();
            }
        }
    }));

    function sendErrorPage (res, flow, err)
    {
        res.content(authorizeTemplate({
            flow: flow,
            error: err
        })).end();
    }
}

function makeToken ()
{
    return StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");
}

function objectHas (object, keys)
{
    return _(object).keys().intersection(keys).value().length === keys.length;
}

function sendError (res, err)
{
    res.status(400).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
        result: "err",
        err: err
    }, true))).end();
}

function sendSuccess (res, data)
{
    res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS(_({}).extend(data).extend({
        result: "ok"
    }).value(), true))).end();
}

function checkForError (error, response, body)
{
    var defaultMessage = "An error occurred.";
    if (error || !body || "err" in body)
    {
        var message = defaultMessage;
        if (error && "javaException" in error)
        {
            error.javaException.printStackTrace();
            if (error.javaException.message)
            {
                message = error.javaException.message.toString();
            }
            else
            {
                message = error.javaException.getClass().getSimpleName().toString();
            }
        }
        else if (!body)
        {
            message = "Empty body";
        }
        else if (typeof body !== "object")
        {
            message = "Unparseable object";
        }
        else if ("err" in body)
        {
            message = body.err;
        }

        if (typeof message === "object")
        {
            message = message.message;
        }

        return message || defaultMessage;
    }
}

function discoverCraniumUrl (flow)
{
    var parsedUserUrl = new URI(flow.userCraniumUrl);
    var portFragment = parsedUserUrl.getPort()  > -1 ? ":" + parsedUserUrl.getPort() : "";
    var possibleUrls = [
        flow.userCraniumUrl,
        "http://" + flow.craniumRequestIp,
        "http://" + flow.craniumRequestIp + portFragment,
        "http://" + flow.craniumRequestIp + parsedUserUrl.getPath(),
        "http://" + flow.craniumRequestIp + portFragment + parsedUserUrl.getPath()
    ];

    logger.info("Attempting to discover Cranium url using candidates: " + possibleUrls.join(", "));

    var promise = null;
    _(possibleUrls).each(function (url) {
        promise = promise ? promise.then(function (discoveredUrl) {
            if (discoveredUrl)
            {
                return discoveredUrl;
            }

            return checkForCranium(url);
        }).catch(function () {
            return checkForCranium(url);
        }) : checkForCranium(url);
    });

    return promise.then(function (discoveredUrl) {
        if (discoveredUrl)
        {
            return discoveredUrl;
        }

        throw new Error("Failed to discover cranium URL");
    });
}

function checkForCranium (url)
{
    logger.info("Attempting to discover Cranium  at " + url);
    return new Promise(function (resolve, reject) {
        request.get(Util.buildUrl(url, "info"), {json: true}, function (err, response, body) {
            var error = checkForError(err, response, body);
            if (error)
            {
                logger.warning("Failed to discover Cranium at " + url + ": " + error);
                reject(new Error(error));
                return;
            }

            logger.info("Discovered Cranium " + body.version + " at " + url);
            resolve(url);
        });
    });
}

function isPreauthorized (/*flow*/)
{
    return false;
}

module.exports = setup;
