/* jshint esnext: true */
var _ = require("lodash");
var request = require("request");
var status = require("lib/status");
var schedule = require("lib/schedule");
var console = require("console")();
var Objects = require("convenience").Objects;

var cJSON = Packages.cattura.utils.parsing.JSON.JSON;
var AtomicBoolean = java.util.concurrent.atomic.AtomicBoolean;
var ExceptionUtils = Packages.cattura.utils.exception.ExceptionUtils;

var DO_NOT_LOG_EVENTS = ["AudioLevelsEvent"];
var http = new (require("lib/http-alt"))();

/*
 * The features object consists of features this version of
 * the Cranium Integration Plugin supports. Cranium can inspect this
 * to determine what features are supported and what versions of features
 * are supported.
 *
 * For versions of Cranium Integration Plugin >= 1.1.0, this should be assumed
 * to be {authv1: 1, internalApi: 1}.
 */
var FEATURES = {
    /*
     * Auth flow v2. This is the "current" auth flow.
     */
    authv2: 1,

    /*
     * The internal API that Cranium and CaptureCast use to talk. This is versioned
     * on both sides.
     */
    internalApi: 1,

    /*
     * Can Cranium authenticate against the external CaptureCast API?
     */
    externalApi: 1
};

function Cranium (data)
{
    var t = _(this).extend(data).value();
    var user = null;

    this.info = {};
    this.available = new AtomicBoolean();
    this.connected = new AtomicBoolean();
    this.asUser = function asUser ()
    {
        var User = Packages.cattura.web.server.interfaces.User;
        
        user = user || new User({
            getUsername: function getUsername ()
            {
                return "Command Center at " + t.url;
            },
            getRoles: function getRoles ()
            {
                return [User.Role.ADMINISTRATOR];
            },
            isRole: function isRole (roleName)
            {
                return roleName == "Administrator";
            },
            can: function can ()
            {
                return true;
            }
        });
        return user;
    };
}

function notAPromise (fn)
{
    var result = {};
    fn(
        value => result.value = value,
        error => result.error = error
    );
    
    if (result.error)
    {
        throw result.error;
    }
    
    return result.value;
}

function notAPromise2 (fn)
{
    var future = com.google.common.util.concurrent.SettableFuture.create();
    fn(
        value => future.set({value: value}),
        error => future.set({error: error})
    );

    var result = future.get();
    if (result.error)
    {
        throw result.error;
    }

    return result.value;
}

function makeAPIRequest (options)
{
    var cranium = options.cranium;
    var endpoint = options.endpoint || "/units/me";
    var method = options.method || "get";
    var data = options.data || {};
    var json = 'json' in options ? options.json : true;

    return notAPromise(function (resolve, reject) {
        request({
            uri: cranium.url + endpoint,
            method: method,
            body: data,
            headers: {
                Authorization: "bearer " + cranium.outboundAccessToken
            },
            json: json,
            client: http,
            executor: null
        }, function (err, response, body) {
            if (cranium.isAvailable() && (err || response.statusCode >= 500))
            {
                cranium.setAvailable(false);
                console.log("Command Center at " + cranium.url + " is no longer available");
            }

            if (err)
            {
                reject(err);
            }
            else if (response.statusCode >= 400)
            {
                reject(new java.lang.Exception("Status code: " + response.statusCode));
            }
            else
            {
                if (!cranium.isAvailable())
                {
                    cranium.setAvailable(true);
                    console.log("Command Center at " + cranium.url + " is now available");
                }

                resolve(body);
            }
        });
    });
}

Cranium.prototype.healthCheck = function healthCheck ()
{
    var t = this;
    console.log("Health checking Command Center at " + this.url);

    try
    {
        makeAPIRequest({
            cranium: this,
            endpoint: "/health",
            method: "get",
            json: false
        });
        t.setAvailable(true);
        return true;        
    }
    catch (err)
    {
        console.log("Health check failed: " + err);
        t.setAvailable(false);
        return false;        
    }
};

Cranium.prototype.getInfo = function getInfo ()
{
    var t = this;
    console.log("Getting info for Command Center at " + this.url);

    try
    {
        var result = makeAPIRequest({
            cranium: this,
            endpoint: "/info",
            method: "get",
            json: true
        });
        t.info = result;
        t.setAvailable(true);
        return result;
    }
    catch (err)
    {
        console.log("getInfo failed: " + err);
        t.setAvailable(false);        
    }
};

Cranium.prototype.isLegacy = function isLegacy ()
{
    return this.info.version && _(this.info.version).startsWith("1.");
};

Cranium.prototype.checkAvailability = function checkAvailability ()
{
    var t = this;
    console.log("Checking availability of Command Center at " + this.url);

    try
    {
        makeAPIRequest({
            cranium: this,
            endpoint: "/units/me",
            method: "get",
            json: false
        });
        t.setAvailable(true);
        return true;
    }
    catch (err)
    {
        console.log("Availability check failed: " + err);
        t.setAvailable(false);
        return false;
    }
};

Cranium.prototype.pushStatusUpdate = function pushStatusUpdate (items)
{
    var data = {
        err: null
    };

    if (!items)
    {
        items = ["info", "network", "storage", "features", "updateCheckStatus"];
    }
    else if (typeof items !== "object")
    {
        items = _(arguments).toArray().value();
    }

    console.log("Pushing status [" + items + "] update to " + this.url);

    if (_(items).includes("info"))
    {
        data.info = status.getUnitInfo();
    }

    if (_(items).includes("network"))
    {
        data.network = status.getNetworkInfo();
    }

    if (_(items).includes("storage"))
    {
        data.storage = status.getDiskSpaceInfo();
    }

    if (_(items).includes("features"))
    {
        data.features = FEATURES;
    }
    
    if (_(items).includes("updateCheckStatus"))
    {
        data.updateCheckStatus = status.getUpdateCheckStatus();
    }    

    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me",
        method: "patch",
        data: data
    });
};

Cranium.prototype.sayHello = function sayHello ()
{
    var t = this;
    var info = t.getInfo();
    t.connected.set(true);
    return info;
};

Cranium.prototype.pushInitialUpdate = function pushInitialUpdate ()
{
    this.pushStatusUpdate();
    this.pushSettingsUpdate();
    this.pushDevicesUpdate();
    this.pushSchedulerConfigUpdate();
    this.pushScheduleUpdate();
    this.pushCaptureSessionsUpdate();
    this.pushTemplatesUpdate();
};

Cranium.prototype.pushDevicesUpdate = function pushDevicesUpdate ()
{
    console.log("Pushing devices update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/devices",
        method: "put",
        data: status.getConnectedDevices()
    });
};

Cranium.prototype.pushTemplatesUpdate = function pushTemplatesUpdate ()
{
    console.log("Pushing templates update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/templates",
        method: "put",
        data: status.getTemplates()
    });
};

Cranium.prototype.pushSettingsUpdate = function pushSettingsUpdate ()
{
    console.log("Pushing settings update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/settings",
        method: "put",
        data: status.getSettings()
    });
};

Cranium.prototype.pushScheduleUpdate = function pushScheduleUpdate ()
{
    console.log("Pushing schedule update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/schedule",
        method: "put",
        data: schedule.getSchedule()
    });
};

Cranium.prototype.pushScheduleEventUpdate = function pushScheduleUpdate (eventObject)
{
    console.log("Pushing single event update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/schedule/" + eventObject.id,
        method: "put",
        data: eventObject
    });
};

Cranium.prototype.pushCaptureSessionsUpdate = function pushCaptureSessionsUpdate ()
{
    console.log("Pushing capture sessions update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/sessions",
        method: "put",
        data: status.getCaptureSessions()
    });
};

Cranium.prototype.pushSchedulerConfigUpdate = function pushSchedulerConfigUpdate ()
{
    console.log("Pushing scheduler config update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/schedulerConfig",
        method: "put",
        data: schedule.getSchedulerConfig()
    });
};

Cranium.prototype.pushMiscUpdate = function pushMiscUpdate (key, data)
{
    console.log("Pushing \"" + key + "\" update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/" + key,
        method: "put",
        data: data
    });
};

Cranium.prototype.pushEvent = function pushEvent (event)
{
    var eventClassName = event.getClass().getSimpleName();
    
    if (!_(DO_NOT_LOG_EVENTS).includes("" + event.type))
    {
        console.log("Pushing " + event.type + " (" + eventClassName + ") event to " + this.url);
    }

    /*
     * Some Java objects can't be serialized with Rhino's JSON implementation
     * so we use ours which is backed by Jackson
     */
    var maybeEventData = cJSON.stringify(event.data);

    if (maybeEventData.hasError())
    {
        console.log("Could not serialize " + event.type + " event: " + maybeEventData.getException().message);
        throw maybeEventData.getException();
    }

    var eventData = JSON.parse(maybeEventData.getVal());

    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/events",
        method: "post",
        data: Objects.toJS({
            type: {
                name: event.type,
                label: event.eventType.label || event.type,
                category: event.eventType.category.name(),
                severity: event.eventType.severity.name()
            },
            className: eventClassName,
            description: event.description,
            time: event.time.time,
            data: eventData
        }, true)
    });
};

Cranium.prototype.pushIncident = function pushEvent (incident)
{
    console.log("Pushing incident event to " + this.url);

    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/incidents",
        method: "post",
        data: new CraniumIncident(incident)
    });
};

Cranium.prototype.setAvailable = function setAvailable (value)
{
    this.available.set(value);
};

Cranium.prototype.isAvailable = function isAvailable ()
{
    return this.available.get();
};

Cranium.prototype.isConnected = function isConnected ()
{
    return this.connected.get();
};

Cranium.prototype.toJSON = function toJSON ()
{
    return _(this).omit(["available"]).value();
};

function CraniumIncident (incident)
{
    var t = _(this).extend(Objects.toJS({
        message: incident.message,
        timestamp: incident.timestamp,
        level: incident.level.toString(),
        internal: incident.internal,
        metadata: JSON.parse(cJSON.stringify(incident.metadata).getVal()),
    }, true)).value();

    var metadata = cJSON.stringify(incident.metadata).getVal();
    if (metadata)
    {
        t.metadata = JSON.parse(metadata);
    }

    if (incident.exception)
    {
        t.exception = Objects.toJS(ExceptionUtils.convertToMap(incident.exception), true);
    }
}

Cranium.FEATURES = FEATURES;

module.exports = Cranium;
