/* jshint undef: true */
/* global require, module, Packages, java, org */

var facade = require("facade");
var Objects = require("convenience").Objects;
var _ = require("lodash");
var jDate = java.util.Date;
var Request = Packages.cattura.agents.configuration.model.vos.Request;

var PluginProxy = Packages.cattura.agents.configuration.model.proxies.PluginProxy;
var SettingsProxy = Packages.cattura.agents.configuration.model.proxies.SettingsProxy;

var SCHEDULER_PLUGIN = "cattura.scheduling.scheduler.SchedulerPlugin";

function isAvailable ()
{
    return !!facade.retrieveProxy("SchedulerProxy");
}

function getThreeDayRange ()
{
    var Interval = org.threeten.extra.Interval;
    var DateTimeUtils = Packages.cattura.scheduling.scheduler.utils.DateTimeUtils;

    return Interval.of(
        DateTimeUtils.intervalFromString("yesterday").getOrRethrow().getStart(),
        DateTimeUtils.intervalFromString("tomorrow").getOrRethrow().getEnd()
    );
}

function getThreeMonthRange ()
{
    var Interval = org.threeten.extra.Interval;
    var DateTimeUtils = Packages.cattura.scheduling.scheduler.utils.DateTimeUtils;

    return Interval.of(
        DateTimeUtils.intervalFromString("last month").getOrRethrow().getStart(),
        DateTimeUtils.intervalFromString("next month").getOrRethrow().getEnd()
    );
}

function getThreeYearRange ()
{
    var Interval = org.threeten.extra.Interval;
    var DateTimeUtils = Packages.cattura.scheduling.scheduler.utils.DateTimeUtils;

    return Interval.of(
        DateTimeUtils.intervalFromString("last year").getOrRethrow().getStart(),
        DateTimeUtils.intervalFromString("next year").getOrRethrow().getEnd()
    );
}

function getMaximumRange ()
{
    var Interval = org.threeten.extra.Interval;
    return Interval.ALL;
}

function getSchedule ()
{
    var schedulerProxy = facade.retrieveProxy("SchedulerProxy");
    var range = getThreeYearRange();

    return {
        range: {
            start: range.getStart().toEpochMilli(),
            end: range.getEnd().toEpochMilli()
        },
        events: _(Objects.toJS(schedulerProxy.getScheduler().getEventsWithAssociatedTasks(
            jDate.from(range.getStart()), jDate.from(range.getEnd())
        ))).map(function (eventWithTask) {
            return createJSEventForEventWithTask(eventWithTask);
        }).value()
    };
}

function updateSchedule (newSchedule)
{
    var schedulerProxy = facade.retrieveProxy("SchedulerProxy");
    var currentSchedule = schedulerProxy.getScheduler().getEventsWithAssociatedTasks(new jDate(newSchedule.range.start), new jDate(newSchedule.range.end));
    var eventsToUpdate = _(newSchedule.events).map(function (eventObject) {
        return createEventWithTask(eventObject);
    }).value();

    /*
     * Find events that need to be deleted or updated.
     */
    /* jshint loopfunc:true */
    for (var i = 0; i < currentSchedule.size(); i++)
    {
        var event = currentSchedule.get(i).getEvent();

        /*
         * This event does not have a corresponding event in the new schedule
         * so we are deleting it.
         */
        if (!_(newSchedule.events).find(function (newEvent) { return newEvent.id == event.getID(); }))
        {
            schedulerProxy.getScheduler().unschedule(currentSchedule.get(i));
        }
    }
    /* jshint loopfunc:false */

    /*
     * Submit updated schedule
     */
    schedulerProxy.getScheduler().schedule(eventsToUpdate);

    return getSchedule();
}

function updateEvent (event)
{
    var eventWithTask = sendRequest(facade.UPDATE_CAPTURE_EVENT, {
        id: event.id,
        presentationTitle: event.title,
        startTime: event.start,
        duration: (event.stop - event.start) / 60000, // Duration is in minutes
        template: event.templateID
    });

    if (eventWithTask)
    {
        return createJSEventForEventWithTask(eventWithTask);
    }
}

function createEvent (event)
{
    var eventWithTask = sendRequest(facade.CREATE_CAPTURE_EVENT, {
        presentationTitle: event.title,
        startTime: event.start,
        duration: (event.stop - event.start) / 60000, // Duration is in minutes
        template: event.templateID
    });

    if (eventWithTask)
    {
        return createJSEventForEventWithTask(eventWithTask);
    }
}

function deleteEvent (event)
{
    var eventWithTask = sendRequest(facade.DELETE_CAPTURE_EVENT, {
        id: event.id
    });

    if (eventWithTask)
    {
        return createJSEventForEventWithTask(eventWithTask);
    }
}

function sendRequest (type, params)
{
    var request = new Request(params);
    facade.sendNotification(type, request);
    return request.get();
}

/*
 * Given JS event object, return an event/task duo.
 */
function createEventWithTask (jsEventObject)
{
    var event = new Packages.cattura.scheduling.scheduler.model.local.LocalEvent();
    var task = new Packages.cattura.scheduling.scheduler.model.tasks.CaptureTask();

    event.setID(jsEventObject.id);
    event.setTitle(jsEventObject.title);
    event.setStartTime(new jDate(jsEventObject.start));
    event.setStopTime(new jDate(jsEventObject.stop));
    event.setGroupID(jsEventObject.group);

    task.setEventID(jsEventObject.id);
    task.setTemplateID(jsEventObject.templateID);
    task.setMediaPackageID(jsEventObject.mediaPackageID);
    return new Packages.cattura.scheduler.model.vos.EventWithTask(event, task);
}

/*
 * Given event/task object, return JS representation used in
 * Cranium and in this plugin
 */
function createJSEventForEventWithTask (event, task)
{
    if (typeof task === "undefined")
    {
        task = event.getTask();
        event = event.getEvent();
    }

    var EventStatus = Packages.cattura.scheduling.scheduler.model.vos.EventStatus;
    var schedulerProxy = facade.retrieveProxy("SchedulerProxy");
    var status = schedulerProxy.getScheduler().getStatusForEvent(event);
    var statusString = status.name().toLowerCase();

    switch (status)
    {
        case EventStatus.FAILED:
            statusString = "error";
            break;
        case EventStatus.SUCCEEDED:
            statusString = "done";
            break;
        case EventStatus.CAPTURING:
            statusString = "current";
            break;
    }

    return Objects.toJS({
        id: event.ID,
        group: event.groupID,
        title: event.title,
        start: event.startTime.time,
        stop: event.stopTime.time,
        templateID: task.templateID,
        mediaPackageID: task.mediaPackageID,
        tags: task.tags,
        status: statusString,
        error: task.errorMessage,
        editable: event.editable,
        sourceID: event.sourceID,
        sourceName: event.sourceName,
        externalEventURL: event.externalEventURL
    }, true);
}

/*
 * Returns the scheduler config, which mostly consists of the configured
 * event source.
 */
function getSchedulerConfig ()
{
    var pluginProxy = facade.retrieveProxy(PluginProxy.NAME);
    var schedulerPlugin = pluginProxy.getPluginInfo(SCHEDULER_PLUGIN).orElse(null);
    if (!schedulerPlugin)
    {
        return {};
    }

    var settingsProxy = facade.retrieveProxy(SettingsProxy.NAME);
    return settingsProxy.getUserSettingsForPluginId(schedulerPlugin.getPluginID()).get().toMap();
}

module.exports = {
    isAvailable: isAvailable,
    getSchedule: getSchedule,
    updateSchedule: updateSchedule,
    updateEvent: updateEvent,
    createEvent: createEvent,
    deleteEvent: deleteEvent,
    createJSEventForEventWithTask: createJSEventForEventWithTask,
    getSchedulerConfig: getSchedulerConfig
};
