/* jshint undef: true, unused: true */
/* global require, com, module, java, Packages */

var facade = require("facade");
var SystemInformationUtilities = Packages.cattura.agents.configuration.model.utilities.SystemInformationUtilities;
var OSSystemState = Packages.cattura.utils.os.OSSystemState;
var TemplateHelper = Packages.cattura.agents.configuration.model.vos.capture.helpers.TemplateHelper;
var MediaTypes = Packages.cattura.utils.media.properties.MediaTypes;
var Processor = com.cattura.share.processor.Processor;
var Objects = require("convenience").Objects;
var _ = require("lodash");

var PluginHelper = require("lib/pluginHelper");
var PluginProxy = Packages.cattura.agents.configuration.model.proxies.PluginProxy;

var LIVE_PREVIEW_PLUGIN = "cattura.web.live_preview.LivePreviewPlugin";
var INSTALLER_INTEGRATION_PLUGIN = "cattura.configuration.installer_integration.InstallerIntegrationPlugin";

module.exports = {
    getUnitInfo: function getUnitInfo ()
    {
        var settingsProxy = facade.retrieveProxy("SettingsProxy");
        var updaterProxy = facade.retrieveProxy("UpdaterProxy");
        var versionInfo = updaterProxy.getVersionInfo();

        var userSettings = settingsProxy.getUserSettings();
        var systemSettings = settingsProxy.getSystemSettings();
        return Objects.toJS({
            unitName: userSettings.getString("unit-name").orElse(null),
            machineName: OSSystemState.getMachineName(),
            uuid: SystemInformationUtilities.getSystemUUID(),
            agentVersion: versionInfo.getCurrentVersion(),
            serial: systemSettings.getString("serial").orElse(null)
        },true);
    },

    getNetworkInfo: function getNetworkInfo ()
    {
        var environment = facade.retrieveProxy("EnvironmentProxy");
        var networkWatcher = environment.networkStatusWatcher;
        return Objects.toJS({
           internet: {
               connected: networkWatcher.hasInternet(),
               ip: networkWatcher.globalIP
           },
           local: {
               ip: networkWatcher.localIP,
               type: networkWatcher.networkType,
               subnet: networkWatcher.subnet,
               gateway: networkWatcher.gateway,
               mac: networkWatcher.getMACAddress()
           }
        },true);
    },

    getDiskSpaceInfo: function getDiskSpaceInfo ()
    {
        var environment = facade.retrieveProxy("EnvironmentProxy");
        var diskSpaceWatcher = environment.diskSpaceWatcher;
        return Objects.toJS({
            total: diskSpaceWatcher.totalSpace,
            usable: diskSpaceWatcher.usableSpace,
            free: diskSpaceWatcher.freeSpace,
            used: diskSpaceWatcher.usedSpace,
            low: diskSpaceWatcher.isDiskSpaceLow()
        },true);
    },

    getConnectedDevices: function getConnectedDevices ()
    {
        var deviceProxy = facade.retrieveProxy("DeviceProxy");
        var pluginProxy = facade.retrieveProxy(PluginProxy.NAME);

        var livePreviewPlugin = pluginProxy.getPluginInfo(LIVE_PREVIEW_PLUGIN).orElse(null);
        var videoDevicePreviewAvailable = false;
        var audioDevicePreviewAvailable = false;
        if (livePreviewPlugin)
        {
            videoDevicePreviewAvailable = true;
            audioDevicePreviewAvailable = true;
        }

        var deviceStructure = [];
        var audioDevices = Objects.toJS(deviceProxy.audioDevices);
        var videoDevices = Objects.toJS(deviceProxy.videoDevices);
        var key;
        var device;

        for (key in audioDevices)
        {
            device = {
                id: audioDevices[key].name,
                name: audioDevices[key].prettyName,
                type: "audio",
                tested: audioDevices[key].tested,
                usable: audioDevices[key].usable,
                usb: audioDevices[key].isUSBDevice()
            };

            if (audioDevicePreviewAvailable)
            {
                device.liveStream = "live/" + audioDevices[key].sanitizedID;
            }

            deviceStructure.push(device);
        }

        for (key in videoDevices)
        {
            device = {
                id: videoDevices[key].name,
                name: videoDevices[key].prettyName,
                type: "video",
                snapshot: "api/1/devices/video/" + videoDevices[key].sanitizedID + "/thumbnail",
                width: videoDevices[key].width,
                height: videoDevices[key].height,
                fps: videoDevices[key].FPS,
                tested: videoDevices[key].tested,
                usable: videoDevices[key].usable
            };

            if (videoDevicePreviewAvailable)
            {
                device.liveStream =  "live/" + videoDevices[key].sanitizedID;
            }

            deviceStructure.push(device);
        }

        return Objects.toJS(deviceStructure, true);
    },

    getCaptureSessions: function getCaptureSessions ()
    {
        var captureStateProxy = facade.retrieveProxy("CaptureStateProxy");
        var pluginProxy = facade.retrieveProxy(PluginProxy.NAME);
        
        var livePreviewPlugin = pluginProxy.getPluginInfo(LIVE_PREVIEW_PLUGIN).orElse(null);
        return _(Objects.toJS(captureStateProxy.captureState.allSessions)).map(function (session) {
            var retval = {
                state: session.state.name(),
                id: session.captureRequest.mediaPackageID,
                title: session.captureRequest.title
            };

            if (session.session)
            {
                _(retval).extend({
                    startTimeMillis: session.session.startTime.time,
                    elapsedMillis: session.session.elapsedTimeMillis,
                }).value();

                if (session.session.stopped)
                {
                    _(retval).extend({
                        stopTimeMillis: session.session.stopTime.time
                    }).value();
                }
            }

            if (livePreviewPlugin)
            {
                var preview = new PluginHelper(livePreviewPlugin).services.getPreviewFor({
                    id: session.captureRequest.mediaPackageID
                });
                
                if (preview)
                {
                    retval.preview = {
                        id: preview.ID,
                        defaultSource: preview.defaultSource.id,
                        sources: _(Objects.toJS(preview.sources)).map(function (source) {
                            var baseUrl = "live/" + preview.ID + "/" + source.id;
                            return {
                                id: source.id,
                                title: source.title,
                                player: baseUrl,
                                dashManifest: "live/dash/" + source.id + ".mpd",
                                thumbnail: source.thumbnail ? baseUrl + "/thumbnail.png"
                                                            : null,
                                image: source.staticImage ? baseUrl + "/image.png"
                                                          : null,
                                hasAudio: source.hasAudio,
                                hasVideo: source.hasVideo
                            };
                        }).value()
                    };
                }
            }

            return Objects.toJS(retval, true);
        }).value();
    },

    getTemplates: function getTemplates ()
    {
        var templates = facade.retrieveProxy("CaptureTemplateProxy");
        var mediaProcessing = facade.retrieveProxy("MediaProcessingProxy").getCaptureModule();
        return _(Objects.toJS(templates.templates)).map(function (template) {
            var helper = new TemplateHelper(template);
            return Objects.toJS({
                id: template.ID,
                title: template.title,
                audioSources: _(Objects.toJS(helper.getSources())).filter(function (source) {
                    return source.type === MediaTypes.Device.AUDIO;
                }).map(function (source) {
                    return {
                        id: source.sourceId
                    };
                }).value(),
                videoSources: _(Objects.toJS(helper.getSources())).filter(function (source) {
                    return source.type === MediaTypes.Device.VIDEO;
                }).map(function (source) {
                    return {
                        id: source.sourceId
                    };
                }).value(),
                destinations: _(Objects.toJS(template.getGraph().selectNodesByType(Processor.TYPE.INTEGRATION))).map(function (node) {
                    var plugin = mediaProcessing.getPlugin(node.pluginID());
                    return {
                        id: node.pluginID(),
                        title: plugin ? plugin.getTitle() : node.getPluginID()
                    };
                }).value()
            }, true);
        }).value();
    },

    getSettings: function getSettings ()
    {
        var settings = facade.retrieveProxy("SettingsProxy");

        return Objects.toJS(settings.getUserSettings().toMap(), true);
    },
    
    getUpdateCheckStatus: function getUpdateCheckStatus ()
    {
        var pluginProxy = facade.retrieveProxy(PluginProxy.NAME);
        
        var installerIntegrationPlugin = pluginProxy.getPluginInfo(INSTALLER_INTEGRATION_PLUGIN).orElse(null);
        if (!installerIntegrationPlugin)
        {
            return {};
        }
        
        return new PluginHelper(installerIntegrationPlugin).services.getFoundUpdates();
    },

    observeCaptureState: function observeCaptureState (fn)
    {
        var captureStateProxy = facade.retrieveProxy("CaptureStateProxy");
        captureStateProxy.getCaptureState().addObserver(new java.util.Observer({
            update: function update (captureState, sessionState)
            {
                fn(captureState, sessionState);
            }
        }));
    }
};
