/* jshint undef: true, unused: true, esnext: true */
/* global plugin, require, com, Packages, org, java */

var _ = require("lodash");
var on = require("on")(plugin);
var Objects = require("convenience").Objects;
var schedule = require("lib/schedule");
var status = require("lib/status");
var Cranium = require("lib/cranium");
var getIPAddress = require("lib/util").getIPAddress;
var EventType = Packages.cattura.agents.configuration.model.vos.events.EventType;
var templates = {
    authorizev1: plugin.resource("templates", "authorize.tmpl"),
    authorizev2: plugin.resource("templates", "authorizev2.tmpl")
};
var STORE = plugin.info().dataStore;
var bearerPrefix = "bearer ";
var facade = require("facade");

var EventHelper = Packages.cattura.agents.configuration.model.utilities.EventHelper;
var IncidentReport = Packages.cattura.agents.configuration.model.vos.IncidentReport;

/*
 * Monitoring and reporting on status of blocked tasks
 */
var FutureStatusWatcher = Packages.cattura.agents.configuration.model.vos.watchers.FutureStatusWatcher;
var Time = Packages.cattura.utils.time.Time;
var Duration = java.time.Duration;
var ChronoUnit = java.time.temporal.ChronoUnit;

/*
 * Logging
 */
var logger = plugin.info().logger;
var console = require("console")(logger);
var Level = java.util.logging.Level;

/*
 * Executors used to handle events and push updates to Cranium
 */
var ThreadFactoryBuilder = com.google.common.util.concurrent.ThreadFactoryBuilder;
var eventExecutors = {};

/*
 * Executor used to spawn other threads as needed
 */
var executor = java.util.concurrent.Executors.newCachedThreadPool(new ThreadFactoryBuilder()
    .setNameFormat("Command Center Integration Plugin executor thread %d")
    .setDaemon(true)
.build());

/*
 * Incidents that we send out. This is to avoid handling our own incidents
 * in our IncidentEvent handler.
 */
var ourIncidents = [];

/*
 * Instances of Cranium that we are connected to
 */
var craniums;

/*
 * Migrate plugin data storage.
 */
function migrateStorage (oldName, newName)
{
    var DataStorageProxy = Packages.cattura.agents.configuration.model.proxies.DataStorageProxy;
    var dataStorageProxy = facade.retrieveProxy(DataStorageProxy.NAME);
    dataStorageProxy.migrate(oldName, newName);
}

migrateStorage("cranium-integration-plugin", "command-center-integration-plugin");

/*
 * Load Craniums from data store
 */
craniums = _(Objects.toJS(STORE.select("craniums[*]")))
    .map(cranium => new Cranium(JSON.parse(cranium)))
    .value();

logger.log(Level.FINE, "Loaded " + craniums.length + " command centers");

/*
 * By default the CIP will broadcast anything that's not an internal event,
 * or is on the allowlist, or is not on the blocklist.
 */
var EVENT_TYPE_ALLOWLIST = [];
var EVENT_TYPE_BLOCKLIST = [];

function eventIsNotifiable (commandCenter, event)
{
    var eventType      = "" + event.type;
    var eventClassName = "" + event.getClass().getSimpleName();
    if (_(EVENT_TYPE_ALLOWLIST).includes(eventType) || _(EVENT_TYPE_ALLOWLIST).includes(eventClassName))
    {
        return true;
    }
    
    if (new EventHelper(event).isInternalEvent()         || // Internal events are kept inside CaptureCast
        event.eventType == EventType.EXCEPTION           || // Exceptions are handled separately
        _(EVENT_TYPE_BLOCKLIST).includes(eventType)      ||
        _(EVENT_TYPE_BLOCKLIST).includes(eventClassName) ||
        (commandCenter.isLegacy() && !event.description))   // Events with no description are nonsense to Command Center v1
    {
        return false;
    }

    return true;
}

/*
 * Register an authorized cranium instance. We will be able to send
 * messages to it, and recieve messages from it.
 */
function addCranium (cranium)
{
    cranium = new Cranium(Objects.toJS(cranium, true));
    cranium.id = craniums.length;
    STORE.set("craniums[" + cranium.id + "]", JSON.stringify({
        id: cranium.id,
        url: cranium.url,
        outboundAccessToken: cranium.outboundAccessToken,
        inboundAccessToken: cranium.inboundAccessToken
    }));
    logger.log(Level.INFO, "Added command center (id: " + cranium.id + ", url: " + cranium.url + ")");
    craniums.push(cranium);
    return cranium;
}

/*
 * Remove a cranium instance from our data store. This will
 * cause us to no longer send that cranium any status updates.
 * This is invoked when the cranium user selects "Remove Unit" from
 * the cranium UI.
 */
function removeCranium (cranium)
{
    STORE.remove("craniums[" + cranium.id + "]");
    logger.log(Level.INFO, "Removed command center (id: " + cranium.id + ", url: " + cranium.url + ")");
    craniums = _(craniums).without(cranium).value();
}

/*
 * Get "available" cranium instances, i.e.
 * those that are online and that we are able
 * to send messages to.
 *
 * A cranium is marked as unavailable if we fail to
 * connect to it, either during initialization or
 * during the lifetime of the plugin. If this happens,
 * the cranium must send a message notifying us that
 * it is available again in order for us to begin
 * sending messages again.
 */
function getAvailableCraniums ()
{
    return _(craniums).filter(cranium => cranium.isAvailable() && cranium.isConnected()).value();
}

/*
 * Filter list of cranium instances such that there is only one instance
 * per url, so that if duplicate entires are accidentally added we
 * don't push to them.
 */
function filterUniqueCraniums (craniums)
{
    return _(craniums).uniqBy("url").value();
}

function getCraniumForRequest (req)
{
    var authorization;
    var token;
    var cranium;

    if (!req.hasHeader("Authorization")) return null;
    authorization = req.header("Authorization");
    if (!authorization.startsWith(bearerPrefix)) return null;

    token = Objects.toJS(authorization.substring(bearerPrefix.length), true);
    cranium = _(craniums).find(cranium => token === cranium.inboundAccessToken);

    if (!cranium)
    {
        logger.log(Level.FINE, "Unable to find command center instance (auth header: " + authorization + ")");
    }

    return cranium;
}

function getEventExecutorForCranium (cranium)
{
    eventExecutors[cranium.url] = eventExecutors[cranium.url] || java.util.concurrent.Executors.newSingleThreadExecutor(new ThreadFactoryBuilder()
        .setNameFormat("Command Center Integration Plugin event pusher thread %d => " + cranium.url)
        .setDaemon(true)
    .build());
    return eventExecutors[cranium.url];
}

function rejectRequest (req, res)
{
    console.log("Rejected request from " + getIPAddress(req) + " due to invalid authorization");
    console.log(" - " + req.header("Authorization"));
    res.status(401).header("Content-Type", "application/json").content(JSON.stringify({err: "Unauthorized"})).end();
}

/*
 * Decorator for internal (Cranium -> CaptureCast) API functions.
 */
function apiHandler (func)
{
    return new org.webbitserver.HttpHandler({
        handleHttpRequest (req, res)
        {
            var cranium = getCraniumForRequest(req);
            if (!cranium)
            {
                rejectRequest(req, res);
                return;
            }

            if (!cranium.isAvailable())
            {
                logger.log(Level.INFO, "Command Center instance at " + cranium.url + " is available");
            }

            cranium.setAvailable(true);
            
            try
            {
                func(req, res, cranium);
            }
            catch (e)
            {
                handleError(e);
                handleAPIError(e, res);
            }
        }
    });
}

function handleError (err, message)
{
    var exception = err ? (err.javaException || new java.lang.Exception(err.message)) : null;
    message = message || "Failed to connect to Command Center";
    var incident = new IncidentReport(message, exception);
    incident.setInternal(true);
    ourIncidents.push(incident);
    facade.sendNotification(facade.HANDLE_EXCEPTION, incident);
}

function handleAPIError (err, res)
{
    res.status(500).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
        result: "err",
        message: err.message
    }, true))).end();
}

function runOnExecutor (title, eventExecutor, fn)
{
    var future = eventExecutor.submit(Objects.toRunnable(() => {
        try
        {
            fn();
        }
        catch (err)
        {
            handleError(err, "Uncaught error in Command Center plugin from task \"" + title + "\"");
        }
    }));

    executor.submit(new FutureStatusWatcher(future, Duration.of(5, ChronoUnit.MINUTES), 5, Objects.toRunnable(() => {
        handleError(null, new IncidentReport("Command Center update thread task \"" + title + "\" timed out"));
        future.cancel(true);
    })));

    return future;
}

/*
 * This thread sends Cranium an update at the start of a new day, to keep
 * schedule updates current.
 */
executor.execute((() => {
    var dateFormatter = new java.text.SimpleDateFormat("y-M-d");
    var currentDate = dateFormatter.format(new java.util.Date());
    var intervalMillis = 1800000; // 30 minutes

    return Objects.toRunnable(() => {
        console.log("Daily update thread starting");

        while (true)
        {
            var nowString = dateFormatter.format(new java.util.Date());
            if (nowString.equalsIgnoreCase(currentDate))
            {
                Time.sleep(intervalMillis);
                continue;
            }

            currentDate = nowString;
            console.log("Pushing out daily update");
            connectCommandCenters();
        }
    });
})());

/*
 * Run a task on the executor for each available cranium (command center) instance.
 */
function runForEachCranium (name, fn, targetCraniums)
{
    targetCraniums = targetCraniums || filterUniqueCraniums(getAvailableCraniums());
    if (targetCraniums.length === 0)
    {
        return false;
    }

    if (typeof fn === "undefined")
    {
        fn = name;
        name = "Command Center update thread task";
    }

    _(targetCraniums).each(cranium => {
        runOnExecutor(
            "[" + cranium.url + "] " + name,
            getEventExecutorForCranium(cranium),
            () => fn(cranium)
        );
    });
    
    return true;
}

/*
 * Push capture session status update. Additionally, if the given mediaPackageIds
 * correspond to scheduled events, those are pushed too. If no mediaPackageIds
 * are given, it will fetch the list of currently running capture sessions and
 * derive mediaPackageIds from that.
 */
function pushCaptureStatusUpdate (mediaPackageIds)
{
    var eventsToUpdate = [];
    var availableCraniums = filterUniqueCraniums(getAvailableCraniums());

    if (availableCraniums.length === 0)
    {
        return;
    }

    if (!mediaPackageIds)
    {
        mediaPackageIds = _(status.getCaptureSessions()).map(function (session) {
            return session.id;
        }).value();
    }

    if (schedule.isAvailable())
    {
        eventsToUpdate = _(schedule.getSchedule().events).filter(function (scheduleEvent) {
            return _(mediaPackageIds).includes(scheduleEvent.mediaPackageID);
        }).value();
    }

    logger.log(Level.FINE, "Pushing capture state update (" + availableCraniums.length + " command centers to be notified)");
    runForEachCranium("Push capture state update", cranium => {
        try
        {
            cranium.pushCaptureSessionsUpdate();
        }
        catch (err)
        {
            handleError(err);
        }

        _(eventsToUpdate).each(event => {
            if (event)
            {
                try
                {
                    cranium.pushScheduleEventUpdate(event);
                }
                catch (err)
                {
                    handleError(err);
                }
            }
        });
    }, availableCraniums);
}

/*
 * Print "connected to" message.
 */
function greet (commandCenter)
{
    console.log("Connected to Command Center (url", commandCenter.url, "version", commandCenter.info.version, ")");    
}

/*
 * For all command centers, check for availability, push status update and set connected flag.
 */
function connectCommandCenters ()
{
    logger.log(Level.FINE, "Pushing initial status updates (" + craniums.length + " command centers to be notified)");
    var validatedUrls = [];
    
    runForEachCranium("Push initial status update", cranium => {
        if (_(validatedUrls).includes(cranium.url))
        {
            console.log("Skipping availability check for duplicate command center", cranium.url);
            return;
        }
        
        const isAvailable = cranium.checkAvailability();
        if (isAvailable)
        {
            validatedUrls.push(cranium.url);
            cranium.sayHello();
            cranium.pushInitialUpdate();
            greet(cranium);
        }
    }, craniums);
}

/*
 * Attach observer to CaptureState object for now, since the CaptureStatusEvent
 * fires too often to be of any practical use (it fires whenever the state of an individual node
 * changes, which happens way too often), whereas the CaptureState observer only fires
 * when the overall state is updated.
 */
status.observeCaptureState(function (captureState) {
    pushCaptureStatusUpdate(_(Objects.toJS(captureState.getAllSessions()))
        .map(sessionState => "" + sessionState.getCaptureRequest().getMediaPackageID())
        .value());
});

on({
    /*
     * Push initial update to Cranium when all Configuration Module
     * plugins are loaded.
     */
    WebServerStartedEvent ()
    {
        connectCommandCenters();
    },

    /*
     * Push general events to Cranium
     */
    Event (event)
    {
        const usableCommandCenters = filterUniqueCraniums(_(getAvailableCraniums())
            .filter(commandCenter => eventIsNotifiable(commandCenter, event))
            .value());
        
        if (usableCommandCenters.length === 0)
        {
            return false;
        }
        
        runForEachCranium(
            "Push event",
            commandCenter => commandCenter.pushEvent(event),
            usableCommandCenters
        );
    },

    /*
     * Push media package related events. This is start/stop/failed etc.
     */
    MediaPackageEvent (event)
    {
        // Don't handle CaptureStatusEvent here
        if (event.getClass().getSimpleName() == "CaptureStatusEvent")
        {
            return;
        }

        if (event.getMediaPackageID())
        {
            pushCaptureStatusUpdate(["" + event.getMediaPackageID()]);
        }
    },

    /*
     * Push capture session update
     * Do not use this event since it fires too often right now.
     */
    /*CaptureStatusEvent: function (event) {

    },*/

    /*
     * Push device detected events to Cranium
     */
    DevicesDetectedEvent ()
    {
        runForEachCranium("Push devices update", cranium => cranium.pushDevicesUpdate());
    },

    /*
     * Push template created/destroy events to Cranium
     */
    TemplateCreatedEvent ()
    {
        runForEachCranium("Push template update", cranium => cranium.pushTemplatesUpdate());
    },

    /*
     * Push device detected events to Cranium
     */
    TemplateDeletedEvent ()
    {
        runForEachCranium("Push template update", cranium => cranium.pushTemplatesUpdate());
    },

    /*
     * Push schedule events to Cranium
     */
    ScheduledEventsEvent ()
    {
        runForEachCranium("Push schedule update", cranium => cranium.pushScheduleUpdate());
    },

    /*
     * Push schedule update and event source update
     */
    SchedulerEventSourceConfiguredEvent ()
    {
        runForEachCranium("Push scheduler configuration update", cranium => {
            cranium.pushSchedulerConfigUpdate();
            cranium.pushScheduleUpdate();
        });
    },

    SchedulerEventSourceSyncEvent ()
    {
        runForEachCranium("Push schedule update", cranium => cranium.pushScheduleUpdate());
    },

    /*
     * When user settings (e.g. label) are updated, push a status update
     * to Cranium to keep the info on that side synced.
     */
    SettingsUpdatedEvent ()
    {
        runForEachCranium("Push update progress update", cranium => {
            cranium.pushStatusUpdate();
            cranium.pushSettingsUpdate();
        });
    },
    
    StorageStatusEvent ()
    {
        runForEachCranium(
            "Push storage status update",
            cranium => cranium.pushStatusUpdate("storage")
        );
    },
    
    NetworkStatusEvent ()
    {
        runForEachCranium(
            "Push network status update",
            cranium => cranium.pushStatusUpdate("network")
        );
    },

    /*
     * Sync update progress to Command Center
     */
    UpdateInstalledEvent (event)
    {
        runForEachCranium(
            "Push update progress update",
            cranium => cranium.pushMiscUpdate("updateProgress", event.data)
        );
    },
    
    /*
     * When updates are finished then clear update progress
     */
    UpdatesInstalledEvent ()
    {
        runForEachCranium(
            "Push update progress update",
            cranium => cranium.pushMiscUpdate("updateProgress", null)
        );
    },

    /*
     * When an incident occurs, notify Cranium.
     */
    IncidentEvent (event)
    {
        var availableCraniums = filterUniqueCraniums(getAvailableCraniums());
        if (availableCraniums.length === 0)
        {
            return;
        }

        var incident = event.getIncident();

        /*
         * If this incident is one we generated, we don't want to handle it here.
         * This is because our incidents are usually due to a failure to connect,
         * so trying to push them can result in an infinite loop.
         */
        if (_(ourIncidents).includes(incident))
        {
            logger.log(Level.FINE, "Not pushing incident \"" + event.incident.message + "\" because it's one of our own");
            return;
        }

        logger.log(Level.FINE, "Pushing incident \"" + event.incident.message + "\" (" + availableCraniums.length + " command centers to be notified)");
        runForEachCranium(
            "Push incident",
            cranium => cranium.pushIncident(incident),
            availableCraniums
        );
    },

    /*
     * Builds the interface for Cranium to communicate with CaptureCast.
     */
    WebServerEvent (event)
    {
        var HttpHandler = org.webbitserver.HttpHandler;
        var server = event.webServer;
        var rest = new org.webbitserver.rest.Rest(server);
        
        var Optional = java.util.Optional;
        
        event.addAuthenticator(Objects.toJavaFunction(req => {
            var cranium = getCraniumForRequest(req);
            if (!cranium)
            {
                return Optional.empty();
            }

            return Optional.of(cranium.asUser());
        }));
        
        /*
         * Used by Cranium to confirm that CaptureCast is compatible with
         * Cranium and can be connected to.
         *
         * The "version" field corresponds to the version of the "internalApi"
         * feature and is deprecated.
         */
        rest.GET("/cranium", new HttpHandler({
            handleHttpRequest (req, res)
            {
                res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                    result: "ok",
                    version: Cranium.FEATURES.internalApi,
                    ip: getIPAddress(req),
                    features: Cranium.FEATURES
                }, true))).end();
            }
        }));

        /*
         * Set up new auth flow.
         */
        require("flow/v2")({
            server: server,
            authorizeTemplate: templates.authorizev2,
            addCranium: addCranium,
            logger: logger,
            greet: greet
        });

        /*
         * Ping endpoint for Cranium, so Cranium can keep check on CaptureCast's
         * connectivity.
         */
        server.add("/cranium/ping", apiHandler((req, res) => {
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({result: "ok"}, true))).end();
        }));

        /*
         * Endpoint for Cranium to request a full update from CaptureCast.
         */
        server.add("/cranium/requestUpdate", apiHandler((req, res, cranium) => {
            logger.fine("Command Center " + cranium.url + " requests update");
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({result: "ok"}, true))).end();
            runOnExecutor("Respond to requestUpdate", getEventExecutorForCranium(cranium), () => {
                cranium.sayHello();
                cranium.pushInitialUpdate();
                greet(cranium);
            });
        }));

        /*
         * Endpoint for Cranium to push a schedule update to CaptureCast.
         * Returns the updated schedule.
         */
        rest.PUT("/cranium/schedule", apiHandler((req, res) => {
            var scheduleData = JSON.parse(req.body());

            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok",
                schedule: schedule.updateSchedule(scheduleData)
            }, true))).end();
        }));

        /*
         * Either creates a new event or performs a batch schedule update.
         */
        rest.POST("/cranium/schedule", apiHandler((req, res) => {
            var scheduleData = JSON.parse(req.body());

            if (scheduleData.range && scheduleData.events)
            {
                res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                    result: "ok",
                    schedule: schedule.updateSchedule(scheduleData)
                }, true))).end();
            }
            else
            {
                res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                    result: "ok",
                    event: schedule.createEvent(scheduleData)
                }, true))).end();
            }
        }));

        /*
         * Endpoint for Cranium to update a single scheduled event.
         * Returns the updated event.
         */
        rest.PUT("/cranium/schedule/{eventId}", apiHandler((req, res) => {
            var eventData = JSON.parse(req.body());
            eventData.id = rest.param(req, "eventId");
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok",
                event: schedule.updateEvent(eventData)
            }, true))).end();
        }));

        /*
         * Endpoint for Cranium to delete a single scheduled event.
         * Returns the updated event.
         */
        rest.DELETE("/cranium/schedule/{eventId}", apiHandler((req, res) => {
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok",
                event: schedule.deleteEvent({
                    id: rest.param(req, "eventId")
                })
            }, true))).end();
        }));

        /*
         * Deletes the record of the requesting Cranium instance.
         * Used when Cranium wants to unlink CaptureCast from itself.
         */
        rest.POST("/cranium/unlink", apiHandler((req, res, cranium) => {
            removeCranium(cranium);
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok"
            }, true))).end();
        }));
    }
});
